import logging
import traceback
from os.path import basename, join

from django import forms
from django.conf import settings
from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.contrib.auth.models import User
from django.contrib.messages import add_message, ERROR
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.db import transaction
from django.http import HttpRequest
from django.urls import ResolverMatch, reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.text import format_lazy, capfirst
from django.utils.translation import ugettext_lazy as _
from jacc.admin import AccountTypeAccountEntryFilter
from jacc.models import Account, EntryType
from jbank.helpers import create_statement, create_reference_payment_batch
from jbank.models import Statement, StatementRecord, StatementRecordSepaInfo, ReferencePaymentRecord, \
    ReferencePaymentBatch, StatementFile, ReferencePaymentBatchFile, Payout, Refund, PayoutStatus, PayoutParty
from jbank.parsers import parse_tiliote_statements, parse_tiliote_statements_from_file, parse_svm_batches_from_file, \
    parse_svm_batches
from jutil.admin import ModelAdminBase


logger = logging.getLogger(__name__)


class SettlementEntryTypesFilter(SimpleListFilter):
    """
    Filters incoming settlement type entries.
    """
    title = _('account entry types')
    parameter_name = 'type'

    def lookups(self, request, model_admin):
        choices = []
        for e in EntryType.objects.all().filter(is_settlement=True).order_by('name'):
            assert isinstance(e, EntryType)
            choices.append((e.id, capfirst(e.name)))

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            return queryset.filter(type__id=val)
        return queryset


class AccountEntryMatchedFilter(SimpleListFilter):
    """
    Filters incoming payments which do not have any child/derived account entries.
    """
    title = _('account.entry.matched.filter')
    parameter_name = 'matched'

    def lookups(self, request, model_admin):
        return [
            ('1', capfirst(_('account.entry.not.matched'))),
            ('2', capfirst(_('account.entry.is.matched'))),
        ]

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            queryset = queryset.filter(type__is_settlement=True, parent=None)
            if val == '1':
                return queryset.filter(child_set=None)
            elif val == '2':
                return queryset.exclude(child_set=None)
        return queryset


class StatementAdmin(ModelAdminBase):
    exclude = ()
    list_per_page = 20
    save_on_top = False
    ordering = ('-record_date', 'account_number')
    date_hierarchy = 'record_date'
    list_filter = (
        'account_number',
    )
    readonly_fields = (
        'file',
        'name',
        'account_number',
        'statement_number',
        'begin_date',
        'end_date',
        'record_date',
        'customer_identifier',
        'begin_balance_date',
        'begin_balance',
        'record_count',
        'currency_code',
        'account_name',
        'account_limit',
        'owner_name',
        'contact_info_1',
        'contact_info_2',
        'bank_specific_info_1',
        'iban',
        'bic',
    )
    search_fields = (
        'name',
        'statement_number',
    )
    list_display = (
        'id',
        'record_date',
        'account_number',
        'statement_number',
        'begin_balance',
        'currency_code',
        'name',
        'account_entry_list'
    )

    def account_entry_list(self, obj):
        assert isinstance(obj, Statement)
        admin_url = reverse('admin:jbank_statementrecord_statement_changelist', args=(obj.id, ))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), StatementRecord.objects.filter(statement=obj).count())
    account_entry_list.short_description = _('account entries')

    def get_urls(self):
        return [
                url(r'^by-file/(?P<file_id>\d+)/$', self.admin_site.admin_view(self.kw_changelist_view), name='jbank_statement_file_changelist'),
            ] + super().get_urls()

    def get_queryset(self, request: HttpRequest):
        rm = request.resolver_match
        assert isinstance(rm, ResolverMatch)
        qs = super().get_queryset(request)
        file_id = rm.kwargs.get('file_id', None)
        if file_id:
            qs = qs.filter(file=file_id)
        return qs


class StatementRecordSepaInfoInlineAdmin(admin.StackedInline):
    exclude = ()
    model = StatementRecordSepaInfo
    can_delete = False
    extra = 0
    readonly_fields = (
        'record',
        'reference',
        'iban_account_number',
        'bic_code',
        'recipient_name_detail',
        'payer_name_detail',
        'identifier',
        'archive_identifier',
    )
    raw_id_fields = (
        'record',
    )


class StatementRecordAdmin(ModelAdminBase):
    exclude = ()
    list_per_page = 50
    save_on_top = False
    date_hierarchy = 'record_date'
    readonly_fields = (
        'id',
        'statement',
        'record_number',
        'archive_identifier',
        'record_date',
        'value_date',
        'paid_date',
        'type',
        'record_code',
        'record_description',
        'amount',
        'receipt_code',
        'delivery_method',
        'name',
        'name_source',
        'recipient_account_number',
        'recipient_account_number_changed',
        'remittance_info',
        'messages',
        'client_messages',
        'bank_messages',
        # from AccountEntry
        'account',
        'timestamp',
        'created',
        'last_modified',
        'type',
        'description',
        'amount',
        'source_file',
        'source_invoice',
        'settled_invoice',
        'settled_item',
        'parent',
    )
    raw_id_fields = (
        'statement',
        # from AccountEntry
        'account',
        'source_file',
        'parent',
        'source_invoice',
        'settled_invoice',
        'settled_item',
    )
    list_filter = (
        AccountEntryMatchedFilter,
        SettlementEntryTypesFilter,
        'record_code',
    )
    search_fields = (
        '=archive_identifier',
        '=amount',
        '=recipient_account_number',
        'record_description',
        'name',
        'remittance_info',
    )
    list_display = (
        'id',
        'record_date',
        'type',
        'record_code',
        'amount',
        'name',
        'recipient_account_number',
        'remittance_info',
        'source_file_link'
    )
    inlines = (
        StatementRecordSepaInfoInlineAdmin,
    )

    def get_urls(self):
        return [
                url(r'^by-statement/(?P<statement_id>\d+)/$', self.admin_site.admin_view(self.kw_changelist_view), name='jbank_statementrecord_statement_changelist'),
            ] + super().get_urls()

    def get_queryset(self, request: HttpRequest):
        rm = request.resolver_match
        assert isinstance(rm, ResolverMatch)
        qs = super().get_queryset(request)
        statement_id = rm.kwargs.get('statement_id', None)
        if statement_id:
            qs = qs.filter(statement__id=statement_id)
        return qs

    def source_file_link(self, obj):
        assert isinstance(obj, StatementRecord)
        if not obj.statement:
            return ''
        admin_url = reverse('admin:jbank_statement_change', args=(obj.statement.id, ))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.statement.name)
    source_file_link.admin_order_field = 'statement'
    source_file_link.short_description = _('account entry source file')


class ReferencePaymentRecordAdmin(ModelAdminBase):
    exclude = ()
    list_per_page = 50
    save_on_top = False
    date_hierarchy = 'record_date'
    raw_id_fields = (
        'batch',
        # from AccountEntry
        'account',
        'source_file',
        'parent',
        'source_invoice',
        'settled_invoice',
        'settled_item',
    )
    readonly_fields = (
        'id',
        'batch',
        'record_type',
        'account_number',
        'record_date',
        'paid_date',
        'archive_identifier',
        'remittance_info',
        'payer_name',
        'currency_identifier',
        'name_source',
        'amount',
        'correction_identifier',
        'delivery_method',
        'receipt_code',
        # from AccountEntry
        'account',
        'timestamp',
        'created',
        'last_modified',
        'type',
        'description',
        'amount',
        'source_file',
        'source_invoice',
        'settled_invoice',
        'settled_item',
        'parent',
    )
    list_filter = (
        AccountEntryMatchedFilter,
        'correction_identifier',
        'receipt_code',
    )
    search_fields = (
        '=archive_identifier',
        '=amount',
        'remittance_info',
        'payer_name',
        'batch__name',
    )
    list_display = (
        'id',
        'record_date',
        'type',
        'amount',
        'payer_name',
        'remittance_info',
        'source_file_link',
    )

    def get_urls(self):
        return [
                url(r'^by-batch/(?P<batch_id>\d+)/$', self.admin_site.admin_view(self.kw_changelist_view), name='jbank_referencepaymentrecord_batch_changelist'),
            ] + super().get_urls()

    def get_queryset(self, request: HttpRequest):
        rm = request.resolver_match
        assert isinstance(rm, ResolverMatch)
        qs = super().get_queryset(request)
        batch_id = rm.kwargs.get('batch_id', None)
        if batch_id:
            qs = qs.filter(batch_id=batch_id)
        return qs

    def source_file_link(self, obj):
        assert isinstance(obj, ReferencePaymentRecord)
        if not obj.batch:
            return ''
        admin_url = reverse('admin:jbank_referencepaymentbatch_change', args=(obj.batch.id, ))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.batch.name)
    source_file_link.admin_order_field = 'batch'
    source_file_link.short_description = _('account entry source file')


class ReferencePaymentBatchAdmin(ModelAdminBase):
    exclude = ()
    list_per_page = 20
    save_on_top = False
    ordering = ('-record_date',)
    date_hierarchy = 'record_date'
    list_filter = (
        'record_set__account_number',
    )
    readonly_fields = (
        'name',
        'record_date',
        'institution_identifier',
        'service_identifier',
        'currency_identifier',
    )
    search_fields = (
        'name',
        '=record_set__archive_identifier',
        '=record_set__amount',
        'record_set__remittance_info',
        'record_set__payer_name',
    )
    list_display = (
        'id',
        'name',
        'record_date',
        'service_identifier',
        'currency_identifier',
        'account_entry_list',
    )

    def account_entry_list(self, obj):
        assert isinstance(obj, ReferencePaymentBatch)
        admin_url = reverse('admin:jbank_referencepaymentrecord_batch_changelist', args=(obj.id, ))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), ReferencePaymentRecord.objects.filter(batch=obj).count())
    account_entry_list.short_description = _('account entries')

    def get_urls(self):
        return [
                url(r'^by-file/(?P<file_id>\d+)/$', self.admin_site.admin_view(self.kw_changelist_view), name='jbank_referencepaymentbatch_file_changelist'),
            ] + super().get_urls()

    def get_queryset(self, request: HttpRequest):
        rm = request.resolver_match
        assert isinstance(rm, ResolverMatch)
        qs = super().get_queryset(request)
        file_id = rm.kwargs.get('file_id', None)
        if file_id:
            qs = qs.filter(file=file_id)
        return qs


class StatementFileForm(forms.ModelForm):
    class Meta:
        exclude = []

    def clean_file(self):
        file = self.cleaned_data['file']
        assert isinstance(file, InMemoryUploadedFile)
        name = file.name
        file.seek(0)
        content = file.read()
        assert isinstance(content, bytes)
        try:
            statements = parse_tiliote_statements(content.decode('ISO-8859-1'), filename=basename(name))
            for stm in statements:
                account_number = stm['header']['account_number']
                if Account.objects.filter(name=account_number).count() == 0:
                    raise ValidationError(_('account.not.found').format(account_number=account_number))
        except ValidationError:
            raise
        except Exception as e:
            raise ValidationError(_('Unhandled error') + ': {}'.format(e))
        return file


class StatementFileAdmin(ModelAdminBase):
    save_on_top = False
    exclude = ()
    form = StatementFileForm

    list_display = (
        'id',
        'created',
        'file_link',
    )

    readonly_fields = (
        'created',
        'errors',
        'file_link',
    )

    def file_link(self, obj):
        assert isinstance(obj, StatementFile)
        if not obj.file:
            return ''
        name = basename(obj.file.name)
        admin_url = reverse('admin:jbank_statement_file_changelist', args=(obj.id, ))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), name)
    file_link.admin_order_field = 'file'
    file_link.short_description = _('statements')

    def construct_change_message(self, request, form, formsets, add=False):
        if add:
            instance = form.instance
            assert isinstance(instance, StatementFile)
            if instance.file:
                full_path = instance.full_path
                plain_filename = basename(full_path)
                try:
                    statements = parse_tiliote_statements_from_file(full_path)
                    with transaction.atomic():
                        for data in statements:
                            create_statement(data, name=plain_filename, file=instance)
                except Exception as e:
                    instance.errors = traceback.format_exc()
                    instance.save()
                    add_message(request, ERROR, str(e))
                    instance.delete()

        return super().construct_change_message(request, form, formsets, add)


class ReferencePaymentBatchFileForm(forms.ModelForm):
    class Meta:
        exclude = []

    def clean_file(self):
        file = self.cleaned_data['file']
        assert isinstance(file, InMemoryUploadedFile)
        name = file.name
        file.seek(0)
        content = file.read()
        assert isinstance(content, bytes)
        try:
            batches = parse_svm_batches(content.decode('ISO-8859-1'), filename=basename(name))
            for b in batches:
                for rec in b['records']:
                    account_number = rec['account_number']
                    if Account.objects.filter(name=account_number).count() == 0:
                        raise ValidationError(_('account.not.found').format(account_number=account_number))
        except ValidationError:
            raise
        except Exception as e:
            raise ValidationError(_('Unhandled error') + ': {}'.format(e))
        return file


class ReferencePaymentBatchFileAdmin(ModelAdminBase):
    save_on_top = False
    exclude = ()
    form = ReferencePaymentBatchFileForm

    list_display = (
        'id',
        'created',
        'file_link',
    )

    search_fields = (
        'file__contains',
    )

    readonly_fields = (
        'created',
        'errors',
        'file_link',
    )

    def file_link(self, obj):
        assert isinstance(obj, ReferencePaymentBatchFile)
        if not obj.file:
            return ''
        name = basename(obj.file.name)
        admin_url = reverse('admin:jbank_referencepaymentbatch_file_changelist', args=(obj.id, ))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), name)
    file_link.admin_order_field = 'file'
    file_link.short_description = _("reference payment batches")

    def construct_change_message(self, request, form, formsets, add=False):
        if add:
            instance = form.instance
            assert isinstance(instance, ReferencePaymentBatchFile)
            if instance.file:
                full_path = instance.full_path
                plain_filename = basename(full_path)
                try:
                    batches = parse_svm_batches_from_file(full_path)
                    with transaction.atomic():
                        for data in batches:
                            create_reference_payment_batch(data, name=plain_filename, file=instance)
                except Exception as e:
                    user = request.user
                    assert isinstance(user, User)
                    instance.errors = traceback.format_exc()
                    instance.save()
                    msg = str(e)
                    if user.is_superuser:
                        msg = instance.errors
                    logger.error('{}: '.format(plain_filename) + msg)
                    add_message(request, ERROR, msg)
                    instance.delete()

        return super().construct_change_message(request, form, formsets, add)


class PayoutStatusInlineAdmin(admin.TabularInline):
    exclude = ()
    model = PayoutStatus
    can_delete = False
    extra = 0
    ordering = ('-id',)
    readonly_fields = (
        'created',
        'file_name',
        'response_code',
        'response_text',
        'msg_id',
        'original_msg_id',
        'group_status',
        'status_reason',
    )


class PayoutAdmin(ModelAdminBase):
    save_on_top = False
    exclude = ()
    inlines = [PayoutStatusInlineAdmin]
    date_hierarchy = 'timestamp'

    raw_id_fields = (
        'account',
        'parent',
        'payer',
        'recipient',
    )

    list_filter = (
        'state',
        'payoutstatus_set__response_code',
        'payoutstatus_set__group_status',
    )

    fields = (
        'account',
        'parent',
        'payer',
        'recipient',
        'amount',
        'messages',
        'reference',
        'msg_id',
        'file_name',
        'timestamp',
        'due_date',
        'paid_date',
        'state',
        'group_status',
        'created',
    )

    list_display = (
        'id',
        'timestamp',
        'recipient',
        'amount',
        'paid_date',
        'state',
    )

    readonly_fields = (
        'created',
        'msg_id',
        'file_name',
        'group_status',
    )

    search_fields = (
        '=msg_id',
        '=file_name',
        '=file_reference',
    )

    def save_model(self, request, obj, form, change):
        assert isinstance(obj, Payout)
        if not change:
            if not hasattr(obj, 'account') or not obj.account:
                obj.account = obj.payer.payouts_account
            if not hasattr(obj, 'type') or not obj.type:
                obj.type = EntryType.objects.get(code=settings.E_BANK_PAYOUT)
        return super().save_model(request, obj, form, change)


class PayoutPartyAdmin(ModelAdminBase):
    save_on_top = False
    exclude = ()

    list_display = (
        'id',
        'name',
        'account_number',
        'bic',
        'org_id',
        'address',
        'country_code',
    )

    raw_id_fields = (
        'payouts_account',
    )


admin.site.register(Payout, PayoutAdmin)
admin.site.register(PayoutParty, PayoutPartyAdmin)
admin.site.register(Refund, PayoutAdmin)
admin.site.register(Statement, StatementAdmin)
admin.site.register(StatementRecord, StatementRecordAdmin)
admin.site.register(StatementFile, StatementFileAdmin)
admin.site.register(ReferencePaymentRecord, ReferencePaymentRecordAdmin)
admin.site.register(ReferencePaymentBatch, ReferencePaymentBatchAdmin)
admin.site.register(ReferencePaymentBatchFile, ReferencePaymentBatchFileAdmin)



from collections import OrderedDict
from datetime import datetime, timezone, date
from xml.etree import ElementTree
from xml.etree.ElementTree import Element
from decimal import Decimal
import pytz
from django.core.exceptions import ValidationError
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from jutil.format import dec2
from jutil.parse import parse_datetime
from jutil.validators import iban_filter, iban_validator, iso_payment_reference_validator, \
    fi_payment_reference_validator
from jutil.xml import dict_to_element, xml_to_dict


PAIN001_REMITTANCE_INFO_MSG = 'M'
PAIN001_REMITTANCE_INFO_OCR = 'O'
PAIN001_REMITTANCE_INFO_OCR_ISO = 'I'

PAIN001_REMITTANCE_INFO_TYPE = (
    (PAIN001_REMITTANCE_INFO_MSG, _('message')),
    (PAIN001_REMITTANCE_INFO_OCR, _('OCR')),
    (PAIN001_REMITTANCE_INFO_OCR_ISO, _('OCR/ISO')),
)

PAIN001_REMITTANCE_INFO_VALUES = [t[0] for t in PAIN001_REMITTANCE_INFO_TYPE]


class Pain001Party(object):
    def __init__(self, name: str, account: str, bic: str, org_id: str='', address_lines: list=list(), country_code: str=''):
        account = iban_filter(account)
        iban_validator(account)
        self.name = name
        self.account = account
        self.bic = bic
        self.org_id = org_id
        self.address_lines = address_lines
        self.country_code = country_code


class Pain001Payment(object):
    def __init__(self, payment_id, creditor: Pain001Party, amount: Decimal, remittance_info: str, remittance_info_type: str, due_date: date):
        self.payment_id = payment_id
        self.creditor = creditor
        self.amount = amount
        self.remittance_info = remittance_info
        self.remittance_info_type = remittance_info_type
        self.due_date = due_date

    def clean(self):
        if not self.remittance_info:
            raise ValidationError(_('pain001.remittance.info.missing'))
        if self.remittance_info_type not in PAIN001_REMITTANCE_INFO_VALUES:
            raise ValidationError(_('pain001.remittance.info.type.invalid'))
        if self.remittance_info_type == PAIN001_REMITTANCE_INFO_MSG:
            if not self.remittance_info:
                raise ValidationError(_('Invalid payment reference: {}').format(self.remittance_info))
        elif self.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR:
            fi_payment_reference_validator(self.remittance_info)
        elif self.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR_ISO:
            iso_payment_reference_validator(self.remittance_info)


class Pain001(object):
    """
    Class for generating pain.001.001.03 SEPA payment XML files.
    """

    pain_element_name = 'CstmrCdtTrfInitn'
    # pain_element_name = 'pain.001.001.03'
    tz_str = 'Europe/Helsinki'
    tz = None

    def __init__(self, msg_id: str,
                 debtor_name: str,
                 debtor_account: str,
                 debtor_bic: str,
                 debtor_org_id: str,
                 debtor_address_lines: list,
                 debtor_country_code: str):
        self.msg_id = msg_id
        self.debtor = Pain001Party(debtor_name, debtor_account, debtor_bic, debtor_org_id, debtor_address_lines, debtor_country_code)
        self.payments = []

    def add_payment(self, payment_id,
                    creditor_name: str,
                    creditor_account: str,
                    creditor_bic: str,
                    amount: Decimal,
                    remittance_info: str,
                    remittance_info_type: str=PAIN001_REMITTANCE_INFO_MSG,
                    due_date: date=None,
                    ):
        if not due_date:
            due_date = self._local_time().date()
        creditor = Pain001Party(creditor_name, creditor_account, creditor_bic)
        p = Pain001Payment(payment_id, creditor, dec2(amount), remittance_info, remittance_info_type, due_date)
        p.clean()
        self.payments.append(p)

    def _ctrl_sum(self) -> Decimal:
        total = Decimal('0.00')
        for p in self.payments:
            assert isinstance(p, Pain001Payment)
            total += p.amount
        return total

    def _append_simple(self, parent: Element, tag: str, value):
        e = Element(tag)
        e.text = str(value)
        parent.append(e)
        return e

    def _local_time(self, t: datetime=None) -> datetime:
        if not t:
            t = now()
        if not self.tz:
            self.tz = pytz.timezone(self.tz_str)
        return t.astimezone(self.tz)

    def _timestamp(self, t: datetime) -> str:
        return self._local_time(t).isoformat()

    def _grp_hdr(self) -> Element:
        g = Element('GrpHdr')
        self._append_simple(g, 'MsgId', self.msg_id)
        self._append_simple(g, 'CreDtTm', self._timestamp(now()))
        self._append_simple(g, 'NbOfTxs', len(self.payments))
        self._append_simple(g, 'CtrlSum', self._ctrl_sum())
        # self._append_simple(g, 'BtchBookg', 'true')  # debit all at once
        # self._append_simple(g, 'Grpg', 'MIXD')
        g.append(dict_to_element({
            'InitgPty': OrderedDict([
                ('Nm', self.debtor.name),
                ('PstlAdr', OrderedDict([
                    ('Ctry', self.debtor.country_code),
                    ('AdrLine', [{'@': l} for l in self.debtor.address_lines]),
                ])),
            ]),
        }))
        return g

    def _pmt_inf(self, p: Pain001Payment) -> Element:
        if p.remittance_info_type == PAIN001_REMITTANCE_INFO_MSG:
            rmt_inf = ['RmtInf', OrderedDict([
                ('Ustrd', p.remittance_info),
            ])]
        elif p.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR:
            rmt_inf = ['RmtInf', OrderedDict([
                ('Strd', OrderedDict([
                    ('CdtrRefInf', OrderedDict([
                        ('Tp', OrderedDict([
                            ('CdOrPrtry', OrderedDict([
                                ('Cd', 'SCOR'),
                            ])),
                        ])),
                        ('Ref', p.remittance_info),
                    ])),
                ])),
            ])]
        elif p.remittance_info_type == PAIN001_REMITTANCE_INFO_OCR_ISO:
            rmt_inf = ['RmtInf', OrderedDict([
                ('Strd', OrderedDict([
                    ('CdtrRefInf', OrderedDict([
                        ('Tp', OrderedDict([
                            ('CdOrPrtry', OrderedDict([
                                ('Cd', 'SCOR'),
                            ])),
                            ('Issr', 'ISO'),
                        ])),
                        ('Ref', p.remittance_info),
                    ])),
                ])),
            ])]
        else:
            raise ValidationError(_('Invalid remittance info type: {}').format(p.remittance_info_type))

        return dict_to_element({
            'PmtInf': OrderedDict([
                ('PmtInfId', str(p.payment_id)),
                ('PmtMtd', 'TRF'),  # payment order
                ('ReqdExctnDt', p.due_date.isoformat()),
                ('Dbtr', OrderedDict([
                    ('Nm', self.debtor.name),
                    ('PstlAdr', OrderedDict([
                        ('Ctry', self.debtor.country_code),
                        ('AdrLine', [{'@': l} for l in self.debtor.address_lines]),
                    ])),
                    ('Id', OrderedDict([
                        ('OrgId', OrderedDict([
                            ('Othr', OrderedDict([
                                ('Id', self.debtor.org_id),
                                ('SchmeNm', OrderedDict([
                                    ('Cd', 'BANK'),
                                ])),
                            ])),
                        ])),
                    ])),
                ])),
                ('DbtrAcct', OrderedDict([
                    ('Id', OrderedDict([
                        ('IBAN', self.debtor.account),
                    ])),
                ])),
                ('DbtrAgt', OrderedDict([
                    ('FinInstnId', OrderedDict([
                        ('BIC', self.debtor.bic),
                    ])),
                ])),
                ('ChrgBr', 'SLEV'),  # FollowingService level
                ('CdtTrfTxInf', OrderedDict([
                    ('PmtId', OrderedDict([
                        ('EndToEndId', str(p.payment_id)),
                    ])),
                    ('Amt', OrderedDict([
                        ('InstdAmt', {'@': str(p.amount), '@Ccy': 'EUR'}),
                    ])),
                    ('UltmtDbtr', OrderedDict([
                        ('Nm', self.debtor.name),
                    ])),
                    ('CdtrAgt', OrderedDict([
                        ('FinInstnId', OrderedDict([
                            ('BIC', p.creditor.bic),
                        ])),
                    ])),
                    ('Cdtr', OrderedDict([
                        ('Nm', p.creditor.name),
                    ])),
                    ('CdtrAcct', OrderedDict([
                        ('Id', OrderedDict([
                            ('IBAN', p.creditor.account),
                        ])),
                    ])),
                    rmt_inf,
                ])),
            ]),
        })

    def render(self) -> bytes:
        if len(self.payments) == 0:
            raise ValidationError('No payments in pain.001.001.03')
        doc = Element('Document', xmlns="urn:iso:std:iso:20022:tech:xsd:pain.001.001.03")
        pain = Element(self.pain_element_name)
        doc.append(pain)
        pain.append(self._grp_hdr())
        for p in self.payments:
            assert isinstance(p, Pain001Payment)
            pain.append(self._pmt_inf(p))
        xml_bytes = ElementTree.tostring(doc, encoding='utf8', method='xml')
        del doc
        return xml_bytes

    def render_to_file(self, filename: str):
        with open(filename, 'wb') as fp:
            fp.write(self.render())


class Pain002(object):
    """
    Class for parsing pain.002.001.03 SEPA payment status XML files.
    """
    credit_datetime = None
    msg_id = None
    original_msg_id = None
    group_status = None
    status_reason = None

    def __init__(self, file_content: bytes):
        self.data = xml_to_dict(file_content)

        rpt = self.data.get('CstmrPmtStsRpt', {})

        grp_hdr = rpt.get('GrpHdr', {})
        credit_datetime_str = grp_hdr.get('CreDtTm')
        self.credit_datetime = parse_datetime(credit_datetime_str)
        self.msg_id = grp_hdr.get('MsgId')

        grp_inf = rpt.get('OrgnlGrpInfAndSts', {})
        self.original_msg_id = grp_inf.get('OrgnlMsgId')
        self.group_status = grp_inf.get('GrpSts')
        self.status_reason = grp_inf.get('StsRsnInf', {}).get('Rsn', {}).get('Prtry', '')

        if not self.msg_id:
            raise ValidationError('MsgId missing')
        if not self.original_msg_id:
            raise ValidationError('OrgnlMsgId missing')
        if not self.group_status:
            raise ValidationError('GrpSts missing')

    def __str__(self):
        return '{}: {} {} {}'.format(self.msg_id, self.original_msg_id, self.group_status, self.status_reason)

    @property
    def is_accepted(self):
        return self.group_status == 'ACCP'

    @property
    def is_rejected(self):
        return self.group_status == 'RJCT'



import logging
from os.path import basename
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from jacc.models import Account, AccountType, EntryType
from jbank.models import Statement, StatementRecord, StatementRecordSepaInfo, ReferencePaymentBatch, \
    ReferencePaymentRecord, StatementFile, ReferencePaymentBatchFile, Payout, PayoutStatus, PAYOUT_PAID


ASSIGNABLE_STATEMENT_HEADER_FIELDS = (
    'account_number',
    'statement_number',
    'begin_date',
    'end_date',
    'record_date',
    'customer_identifier',
    'begin_balance_date',
    'begin_balance',
    'record_count',
    'currency_code',
    'account_name',
    'account_limit',
    'owner_name',
    'contact_info_1',
    'contact_info_2',
    'bank_specific_info_1',
    'iban',
    'bic',
)

ASSIGNABLE_STATEMENT_RECORD_FIELDS = (
    'record_number',
    'archive_identifier',
    'record_date',
    'value_date',
    'paid_date',
    'entry_type',
    'record_code',
    'record_description',
    'amount',
    'receipt_code',
    'delivery_method',
    'name',
    'name_source',
    'recipient_account_number',
    'recipient_account_number_changed',
    'remittance_info',
)

ASSIGNABLE_STATEMENT_RECORD_SEPA_INFO_FIELDS = (
    'reference',
    'iban_account_number',
    'bic_code',
    'recipient_name_detail',
    'payer_name_detail',
    'identifier',
    'archive_identifier',
)

ASSIGNABLE_REFERENCE_PAYMENT_BATCH_HEADER_FIELDS = (
    'record_date',
    'institution_identifier',
    'service_identifier',
    'currency_identifier',
)

ASSIGNABLE_REFERENCE_PAYMENT_RECORD_FIELDS = (
    'record_type',
    'account_number',
    'record_date',
    'paid_date',
    'archive_identifier',
    'remittance_info',
    'payer_name',
    'currency_identifier',
    'name_source',
    'amount',
    'correction_identifier',
    'delivery_method',
    'receipt_code',
)

logger = logging.getLogger(__name__)


@transaction.atomic
def create_statement(statement_data: dict, name: str, file: StatementFile, **kw) -> Statement:
    """
    Creates Statement from statement data parsed by parse_tiliote_statements()
    :param statement_data: See parse_tiliote_statements
    :param name: File name of the account statement
    :param file: Source statement file
    :return: Statement
    """
    if 'header' not in statement_data or not statement_data['header']:
        raise ValidationError('Invalid header field in statement data {}: {}'.format(name, statement_data.get('header')))
    header = statement_data['header']

    account_number = header['account_number']
    if not account_number:
        raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=''))
    accounts = list(Account.objects.filter(name=account_number))
    if len(accounts) != 1:
        raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=account_number))
    account = accounts[0]
    assert isinstance(account, Account)

    if Statement.objects.filter(name=name, account=account).first():
        raise ValidationError('Bank account {} statement {} of processed already'.format(account_number, name))
    stm = Statement(name=name, account=account, file=file)
    for k in ASSIGNABLE_STATEMENT_HEADER_FIELDS:
        if k in header:
            setattr(stm, k, header[k])
    # pprint(statement_data['header'])
    for k, v in kw.items():
        setattr(stm, k, v)
    stm.full_clean()
    stm.save()

    if EntryType.objects.filter(code=settings.E_BANK_DEPOSIT).count() == 0:
        raise ValidationError(_('entry.type.missing') + ' ({}): {}'.format('settings.E_BANK_DEPOSIT', settings.E_BANK_DEPOSIT))
    if EntryType.objects.filter(code=settings.E_BANK_WITHDRAW).count() == 0:
        raise ValidationError(_('entry.type.missing') + ' ({}): {}'.format('settings.E_BANK_WITHDRAW', settings.E_BANK_WITHDRAW))
    entry_types = {
        '1': EntryType.objects.get(code=settings.E_BANK_DEPOSIT),
        '2': EntryType.objects.get(code=settings.E_BANK_WITHDRAW),
    }

    for rec_data in statement_data['records']:
        e_type = entry_types.get(rec_data['entry_type'])
        rec = StatementRecord(statement=stm, account=account, type=e_type)
        for k in ASSIGNABLE_STATEMENT_RECORD_FIELDS:
            if k in rec_data:
                setattr(rec, k, rec_data[k])
        for k in ('messages', 'client_messages', 'bank_messages'):
            if k in rec_data:
                setattr(rec, k, '\n'.join(rec_data[k]))
        rec.full_clean()
        rec.save()

        if 'sepa' in rec_data:
            sepa_info_data = rec_data['sepa']
            sepa_info = StatementRecordSepaInfo(record=rec)
            for k in ASSIGNABLE_STATEMENT_RECORD_SEPA_INFO_FIELDS:
                if k in sepa_info_data:
                    setattr(sepa_info, k, sepa_info_data[k])
            # pprint(rec_data['sepa'])
            sepa_info.full_clean()
            sepa_info.save()


@transaction.atomic
def create_reference_payment_batch(batch_data: dict, name: str, file: ReferencePaymentBatchFile, **kw) -> ReferencePaymentBatch:
    """
    Creates ReferencePaymentBatch from data parsed by parse_svm_batches()
    :param batch_data: See parse_svm_batches
    :param name: File name of the batch file
    :return: ReferencePaymentBatch
    """
    if ReferencePaymentBatch.objects.exclude(file=file).filter(name=name).first():
        raise ValidationError('Reference payment batch file {} already exists'.format(name))

    if 'header' not in batch_data or not batch_data['header']:
        raise ValidationError('Invalid header field in reference payment batch data {}: {}'.format(name, batch_data.get('header')))
    header = batch_data['header']

    batch = ReferencePaymentBatch(name=name, file=file)
    for k in ASSIGNABLE_REFERENCE_PAYMENT_BATCH_HEADER_FIELDS:
        if k in header:
            setattr(batch, k, header[k])
    # pprint(statement_data['header'])
    for k, v in kw.items():
        setattr(batch, k, v)
    batch.full_clean()
    batch.save()
    e_type = EntryType.objects.get(code=settings.E_BANK_REFERENCE_PAYMENT)

    for rec_data in batch_data['records']:
        account_number = rec_data['account_number']
        if not account_number:
            raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=''))
        accounts = list(Account.objects.filter(name=account_number))
        if len(accounts) != 1:
            raise ValidationError('{name}: '.format(name=name) + _("account.not.found").format(account_number=account_number))
        account = accounts[0]
        assert isinstance(account, Account)

        rec = ReferencePaymentRecord(batch=batch, account=account, type=e_type)
        for k in ASSIGNABLE_REFERENCE_PAYMENT_RECORD_FIELDS:
            if k in rec_data:
                setattr(rec, k, rec_data[k])
        # pprint(rec_data)
        rec.full_clean()
        rec.save()


def get_or_create_bank_account(account_number: str, currency: str='EUR') -> Account:
    a_type, created = AccountType.objects.get_or_create(code=settings.ACCOUNT_BANK_ACCOUNT, is_asset=True, defaults={'name': _('bank account')})
    acc, created = Account.objects.get_or_create(name=account_number, type=a_type, currency=currency)
    return acc


def process_pain002_file_content(bcontent: bytes, filename: str):
    from jbank.sepa import Pain002

    s = Pain002(bcontent)
    p = Payout.objects.filter(msg_id=s.original_msg_id).first()
    ps = PayoutStatus(payout=p, file_name=basename(filename), msg_id=s.msg_id, original_msg_id=s.original_msg_id, group_status=s.group_status, status_reason=s.status_reason[:255])
    ps.full_clean()
    fields = (
        'payout',
        'file_name',
        'response_code',
        'response_text',
        'msg_id',
        'original_msg_id',
        'group_status',
        'status_reason',
    )
    params = {}
    for k in fields:
        params[k] = getattr(ps, k)
    ps_old = PayoutStatus.objects.filter(**params).first()
    if ps_old:
        ps = ps_old
    else:
        ps.save()
        logger.info('{} status updated {}'.format(p, ps))
    if p:
        if ps.is_accepted:
            p.state = PAYOUT_PAID
            p.paid_date = s.credit_datetime
            p.save(update_fields=['state', 'paid_date'])
            logger.info('{} marked as paid {}'.format(p, ps))
    return ps



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-11-02 03:38
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0006_account_name'),
        ('jbank', '0002_auto_20171031_0356'),
    ]

    operations = [
        migrations.AddField(
            model_name='statement',
            name='account',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='jacc.Account'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.0 on 2018-01-26 08:51

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jutil.validators


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0008_auto_20180121_1705'),
        ('jbank', '0010_auto_20171226_1013'),
    ]

    operations = [
        migrations.CreateModel(
            name='Refund',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('iban_account_number', models.CharField(max_length=35, validators=[jutil.validators.iban_validator], verbose_name='IBAN')),
                ('refund_date', models.DateField(blank=True, db_index=True, default=None, null=True, verbose_name='refund date')),
                ('archive_identifier', models.CharField(blank=True, db_index=True, default='', max_length=32, verbose_name='archive identifier')),
                ('state', models.CharField(blank=True, db_index=True, default='W', max_length=1, verbose_name='state')),
                ('source', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='jacc.AccountEntry')),
            ],
            options={
                'verbose_name_plural': 'refunds',
                'verbose_name': 'refund',
            },
        ),
        migrations.AlterField(
            model_name='referencepaymentbatchfile',
            name='created',
            field=models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
        migrations.AlterField(
            model_name='statementfile',
            name='created',
            field=models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
    ]



# Generated by Django 2.0 on 2017-12-26 10:13

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0009_auto_20171107_1847'),
    ]

    operations = [
        migrations.AlterField(
            model_name='statement',
            name='account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='jacc.Account'),
        ),
    ]



# Generated by Django 2.0.2 on 2018-04-19 13:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0022_auto_20180411_1814'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payout',
            name='messages',
            field=models.TextField(verbose_name='recipient messages'),
        ),
    ]



import logging
import re
from copy import copy
from os.path import basename
from pprint import pprint
from datetime import time, datetime, date
from decimal import Decimal
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext as _
from pytz import timezone


REGEX_SIMPLE_FIELD = re.compile(r'^(X|9)+$')
REGEX_VARIABLE_FIELD = re.compile(r'^(X|9)\((\d+)\)$')

TO_FILE_HEADER_TYPES = ('T00', )

TO_FILE_HEADER_DATES = (
    'begin_date',
    'end_date',
    ('record_date', 'record_time'),
    'begin_balance_date',
)

TO_FILE_HEADER_DECIMALS = (
    ('begin_balance', 'begin_balance_sign'),
    'account_limit',
)

TO_FILE_HEADER = (
    ('statement_type', 'X', 'P'),
    ('record_type', 'XX', 'P'),
    ('record_length', '9(3)', 'P'),
    ('version', 'X(3)', 'P'),
    ('account_number', 'X(14)', 'P'),
    ('statement_number', '9(3)', 'P'),
    ('begin_date', '9(6)', 'P'),
    ('end_date', '9(6)', 'P'),
    ('record_date', '9(6)', 'P'),
    ('record_time', '9(4)', 'P'),
    ('customer_identifier', 'X(17)', 'P'),
    ('begin_balance_date', '9(6)', 'P'),
    ('begin_balance_sign', 'X', 'P'),
    ('begin_balance', '9(18)', 'P'),
    ('record_count', '9(6)', 'P'),
    ('currency_code', 'X(3)', 'P'),
    ('account_name', 'X(30)', 'V'),
    ('account_limit', '9(18)', 'P'),
    ('owner_name', 'X(35)', 'P'),
    ('contact_info_1', 'X(40)', 'P'),
    ('contact_info_2', 'X(40)', 'V'),
    ('bank_specific_info_1', 'X(30)', 'V'),
    ('iban_and_bic', 'X(30)', 'V'),
)

TO_FILE_RECORD_TYPES = ('T10', 'T80')

TO_FILE_RECORD_DATES = (
    'record_date',
    'value_date',
    'paid_date',
)

TO_FILE_RECORD_DECIMALS = (
    ('amount', 'amount_sign'),
)

TO_FILE_RECORD = (
    ('statement_type', 'X', 'P'),
    ('record_type', 'XX', 'P'),
    ('record_length', '9(3)', 'P'),
    ('record_number', '9(6)', 'P'),
    ('archive_identifier', 'X(18)', 'V'),
    ('record_date', '9(6)', 'P'),
    ('value_date', '9(6)', 'V'),
    ('paid_date', '9(6)', 'V'),
    ('entry_type', 'X', 'P'),  # 1 = pano, 2 = otto, 3 = panon korjaus, 4 = oton korjaus, 9 = hylätty tapahtuma
    ('record_code', 'X(3)', 'P'),
    ('record_description', 'X(35)', 'P'),
    ('amount_sign', 'X', 'P'),
    ('amount', '9(18)', 'P'),
    ('receipt_code', 'X', 'P'),
    ('delivery_method', 'X', 'P'),
    ('name', 'X(35)', 'V'),
    ('name_source', 'X', 'V'),
    ('recipient_account_number', 'X(14)', 'V'),
    ('recipient_account_number_changed', 'X', 'V'),
    ('remittance_info', 'X(20)', 'V'),
    ('form_number', 'X(8)', 'V'),
    ('level_identifier', 'X', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_TYPES = ('T11', 'T81')

TO_FILE_RECORD_EXTRA_INFO_HEADER = (
    ('statement_type', 'X', 'P'),
    ('record_type', 'XX', 'P'),
    ('record_length', '9(3)', 'P'),
    ('extra_info_type', '9(2)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_COUNTS = (
    ('entry_count', '9(8)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_INVOICE = (
    ('customer_number', 'X(10)', 'P'),
    ('pad01', 'X', 'P'),
    ('invoice_number', 'X(15)', 'P'),
    ('pad02', 'X', 'P'),
    ('invoice_date', 'X(6)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_CARD = (
    ('card_number', 'X(19)', 'P'),
    ('pad01', 'X', 'P'),
    ('merchant_reference', 'X(14)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_CORRECTION = (
    ('original_archive_identifier', 'X(18)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_CURRENCY_DECIMALS = (
    ('amount', 'amount_sign'),
)

TO_FILE_RECORD_EXTRA_INFO_CURRENCY = (
    ('amount_sign', 'X', 'P'),
    ('amount', '9(18)', 'P'),
    ('pad01', 'X', 'P'),
    ('currency_code', 'X(3)', 'P'),
    ('pad02', 'X', 'P'),
    ('exchange_rate', '9(11)', 'P'),
    ('rate_reference', 'X(6)', 'V'),
)

TO_FILE_RECORD_EXTRA_INFO_REASON = (
    ('reason_code', '9(3)', 'P'),
    ('pad01', 'X', 'P'),
    ('reason_description', 'X(31)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_NAME_DETAIL = (
    ('name_detail', 'X(35)', 'P'),
)

TO_FILE_RECORD_EXTRA_INFO_SEPA = (
    ('reference', 'X(35)', 'V'),
    ('iban_account_number', 'X(35)', 'V'),
    ('bic_code', 'X(35)', 'V'),
    ('recipient_name_detail', 'X(70)', 'V'),
    ('payer_name_detail', 'X(70)', 'V'),
    ('identifier', 'X(35)', 'V'),
    ('archive_identifier', 'X(35)', 'V'),
)

TO_BALANCE_RECORD_DATES = (
    'record_date',
)

TO_BALANCE_RECORD_DECIMALS = (
    ('end_balance', 'end_balance_sign'),
    ('available_balance', 'available_balance_sign'),
)

TO_BALANCE_RECORD = (
    ('statement_type', 'X', 'P'),
    ('record_type', 'XX', 'P'),
    ('record_length', '9(3)', 'P'),
    ('record_date', '9(6)', 'P'),
    ('end_balance_sign', 'X', 'P'),
    ('end_balance', '9(18)', 'P'),
    ('available_balance_sign', 'X', 'P'),
    ('available_balance', '9(18)', 'P'),
)

TO_CUMULATIVE_RECORD_DATES = (
    'period_date',
)

TO_CUMULATIVE_RECORD_DECIMALS = (
    ('deposits_amount', 'deposits_sign'),
    ('withdrawals_amount', 'withdrawals_sign'),
)

TO_CUMULATIVE_RECORD = (
    ('statement_type', 'X', 'P'),
    ('record_type', 'XX', 'P'),
    ('record_length', '9(3)', 'P'),
    ('period_identifier', 'X', 'P'),  # 1=day, 2=term, 3=month, 4=year
    ('period_date', '9(6)', 'P'),
    ('deposits_count', '9(8)', 'P'),
    ('deposits_sign', 'X', 'P'),
    ('deposits_amount', '9(18)', 'P'),
    ('withdrawals_count', '9(8)', 'P'),
    ('withdrawals_sign', 'X', 'P'),
    ('withdrawals_amount', '9(18)', 'P'),
)

TO_SPECIAL_RECORD = (
    ('statement_type', 'X', 'P'),
    ('record_type', 'XX', 'P'),
    ('record_length', '9(3)', 'P'),
    ('bank_group_identifier', 'X(3)', 'P'),
)

SVM_FILE_HEADER_DATES = (
    ('record_date', 'record_time'),
)

SVM_FILE_HEADER_TYPES = (
    '0',
)

SVM_FILE_HEADER = (
    ('statement_type', '9(1)', 'P'),
    ('record_date', '9(6)', 'P'),
    ('record_time', '9(4)', 'P'),
    ('institution_identifier', 'X(2)', 'P'),
    ('service_identifier', 'X(9)', 'P'),
    ('currency_identifier', 'X(1)', 'P'),
    ('pad01', 'X(67)', 'P'),
)

SVM_FILE_RECORD_TYPES = ('3', '5')

SVM_FILE_RECORD_DECIMALS = (
    'amount',
)

SVM_FILE_RECORD_DATES = (
    'record_date',
    'paid_date',
)

SVM_FILE_RECORD = (
    ('record_type', '9(1)', 'P'),  # 3=viitesiirto, 5=suoraveloitus
    ('account_number', '9(14)', 'P'),
    ('record_date', '9(6)', 'P'),
    ('paid_date', '9(6)', 'P'),
    ('archive_identifier', 'X(16)', 'P'),
    ('remittance_info', '9(20)', 'P'),
    ('payer_name', 'X(12)', 'P'),
    ('currency_identifier', 'X(1)', 'P'),  # 1=eur
    ('name_source', 'X', 'V'),
    ('amount', '9(10)', 'P'),
    ('correction_identifier', 'X', 'V'),  # 0=normal, 1=correction
    ('delivery_method', 'X', 'P'),  # A=asiakkaalta, K=konttorista, J=pankin jarjestelmasta
    ('receipt_code', 'X', 'P'),
)

SVM_FILE_SUMMARY_TYPES = (
    '9',
)

SVM_FILE_SUMMARY_DECIMALS = (
    'record_amount',
    'correction_amount',
)

SVM_FILE_SUMMARY = (
    ('record_type', '9(1)', 'P'),  # 9
    ('record_count', '9(6)', 'P'),
    ('record_amount', '9(11)', 'P'),
    ('correction_count', '9(6)', 'P'),
    ('correction_amount', '9(11)', 'P'),
    ('pad01', 'X(5)', 'P'),
)

logger = logging.getLogger(__name__)


def parse_record_format(fmt: str) -> tuple:
    """
    :param fmt: Data format used in .TO files
    :return: Data type ('X' or '9'), data length (number of characters)
    """
    res = REGEX_SIMPLE_FIELD.match(fmt)
    data_type, data_len = None, None
    if res:
        data_type = res.group(1)
        data_len = len(fmt)
    else:
        res = REGEX_VARIABLE_FIELD.match(fmt)
        if res:
            data_type = res.group(1)
            data_len = int(res.group(2))
    if not data_type or not data_len:
        raise Exception('Failed to parse data format {}'.format(fmt))
    return data_type, data_len


def parse_record_value(data_type, data_len, data, name: str, line_number: int) -> str:
    value = data[:data_len]
    if len(value) != data_len:
        raise ValidationError(_('Line {line}: Invalid field "{field}" value "{value}"').format(line=line_number, field=name, value=value))
    if data_type == 'X':
        pass
    elif data_type == '9':
        charset = '0123456789'
        for ch in value:
            if ch not in charset:
                raise ValidationError(_('Line {line}: Invalid field "{field}" value "{value}"').format(line=line_number, field=name, value=value))
        # logger.info('jbank.parsers.parse_record_value: {} = {}'.format(name, value))
    else:
        raise ValidationError(_('Line {line}: Invalid field "{field}" value "{value}"').format(line=line_number, field=name, value=value))
    return value


def parse_records(line: str, specs: tuple, line_number: int, check_record_length: bool=True, record_length: int=None) -> dict:
    # print(line)
    i = 0
    data = {}
    for name, fmt, req in specs:
        data_type, data_len = parse_record_format(fmt)
        value = parse_record_value(data_type, data_len, line[i:], name=name, line_number=line_number)
        # print('[{}:{}] {}="{}"'.format(i, i+data_len, name, value))
        data[name] = str(value).strip()
        i += data_len
    data['extra_data'] = line[i:]

    if 'record_length' in data:
        record_length = data['record_length']
    if check_record_length and record_length:
        data['extra_data'] = data['extra_data'].strip()
        if i != record_length and data['extra_data'] != '':
            raise ValidationError(_('Line {line}: Record length ({record_length}) does not match length of parsed data ({data_length}). Extra data: "{extra_data}"').format(line=line_number, data_length=i+len(data['extra_data']), record_length=record_length, extra_data=data['extra_data']))
    return data


def parse_record_messages(extra_data: str) -> str:
    msg = []
    while extra_data:
        msg.append(extra_data[:35])
        extra_data = extra_data[35:]
    return msg


def parse_record_extra_info(record: dict, line: str, line_number: int) -> dict:
    assert line[:3] in TO_FILE_RECORD_EXTRA_INFO_TYPES

    header = parse_records(line, TO_FILE_RECORD_EXTRA_INFO_HEADER, line_number, check_record_length=False)
    extra_info_type = header['extra_info_type']
    # print(line)
    extra_data = copy(header['extra_data'])
    if extra_info_type == '00':
        record['messages'] = parse_record_messages(extra_data)
    elif extra_info_type == '01':
        record['counts'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_COUNTS, line_number, record_length=8)
    elif extra_info_type == '02':
        record['invoice'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_INVOICE, line_number, record_length=33)
    elif extra_info_type == '03':
        record['card'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_CARD, line_number, record_length=34)
    elif extra_info_type == '04':
        record['correction'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_CORRECTION, line_number, record_length=18)
    elif extra_info_type == '05':
        record['currency'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_CURRENCY, line_number, record_length=41)
        convert_decimal_fields(record['currency'], TO_FILE_RECORD_EXTRA_INFO_CURRENCY_DECIMALS)
    elif extra_info_type == '06':
        record['client_messages'] = parse_record_messages(extra_data)
    elif extra_info_type == '07':
        record['bank_messages'] = parse_record_messages(extra_data)
    elif extra_info_type == '08':
        record['reason'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_REASON, line_number, record_length=35)
    elif extra_info_type == '09':
        record['name_detail'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_NAME_DETAIL, line_number, record_length=35)
    elif extra_info_type == '11':
        record['sepa'] = parse_records(extra_data, TO_FILE_RECORD_EXTRA_INFO_SEPA, line_number, record_length=323)
    else:
        raise ValidationError(_('Line {line}: Invalid record extra info type "{extra_info_type}"').format(line=line_number, extra_info_type=extra_info_type))


def convert_date(v: str, field_name: str) -> date:
    if v is None:
        raise ValidationError(_("Date field missing: {}").format(field_name))
    if len(v) != 6:
        raise ValidationError(_("Date format error in field {}: {}").format(field_name, v))
    year = int(v[0:2]) + 2000
    month = int(v[2:4])
    day = int(v[4:6])
    return date(year=year, month=month, day=day)


def convert_time(v: str, field_name: str) -> time:
    if v is None:
        raise ValidationError(_("Time field missing: {}").format(field_name))
    if not re.match(r'^\d\d\d\d$', v):
        raise ValidationError(_("Time format error in field {}: {}").format(field_name, v))
    return time(int(v[0:2]), int(v[2:4]))


def convert_date_fields(data: dict, date_fields: tuple, tz: timezone):
    for k in date_fields:
        if isinstance(k, str):
            v_date = data.get(k)
            if v_date:
                data[k] = convert_date(v_date, k)
        elif isinstance(k, tuple):
            if len(k) != 2:
                raise ValidationError(_("Date format error in field {}: {}").format(k, v))
            k_date, k_time = k
            v_date, v_time = data.get(k_date), data.get(k_time)
            # print('Converting {}'.format(k))
            # pprint(data)
            if v_date or v_time:
                v_date = convert_date(v_date, k_date)
                v_time = convert_time(v_time, k_time)
                v_datetime = datetime.combine(v_date, v_time)
                data[k_date] = tz.localize(v_datetime)
                del data[k_time]


def convert_decimal_fields(data: dict, decimal_fields: tuple):
    for field in decimal_fields:
        if isinstance(field, str):
            v_number = data.get(field)
            if v_number is not None:
                v = Decimal(v_number) * Decimal('0.01')
                # logger.info('jbank.parsers.convert_decimal_fields: {} = {}'.format(field, v))
                data[field] = v
        elif isinstance(field, tuple) and len(field) == 2:
            k_number, k_sign = field
            v_number, v_sign = data.get(k_number), data.get(k_sign)
            if v_number is not None:
                v = Decimal(v_number) * Decimal('0.01')
                if v_sign == '-':
                    v = -v
                data[k_number] = v
                # logger.info('jbank.parsers.convert_decimal_fields: {} = {}'.format(k_number, v))
                del data[k_sign]
        else:
            raise ValidationError(_('Invalid decimal field format: {}').format(field))


def combine_statement(header, records, balance, cumulative, cumulative_adjustment, special_records) -> dict:
    data = {
        'header': header,
        'records': records,
    }
    if balance is not None:
        data['balance'] = balance
    if cumulative is not None:
        data['cumulative'] = cumulative
    if cumulative_adjustment is not None:
        data['cumulative_adjustment'] = cumulative_adjustment
    if special_records:
        data['special_records'] = special_records
    return data


def parse_tiliote_statements(content: str, filename: str) -> list:
    lines = content.split('\n')
    nlines = len(lines)

    line_number = 1
    tz = timezone('Europe/Helsinki')

    header = None
    records = []
    balance = None
    cumulative = None
    cumulative_adjustment = None
    special_records = []
    statements = []

    while line_number <= nlines:
        line = lines[line_number-1]
        if line.strip() == '':
            line_number += 1
            continue
        # print('parsing line {}: {}'.format(line_number, line))
        record_type = line[:3]

        if record_type in TO_FILE_HEADER_TYPES:
            if header:
                statements.append(combine_statement(header, records, balance, cumulative, cumulative_adjustment, special_records))
                header, records, balance, cumulative, cumulative_adjustment, special_records = None, [], None, None, None, []

            header = parse_records(lines[line_number - 1], TO_FILE_HEADER, line_number=line_number)
            convert_date_fields(header, TO_FILE_HEADER_DATES, tz)
            convert_decimal_fields(header, TO_FILE_HEADER_DECIMALS)
            iban_and_bic = header.get('iban_and_bic', '').split(' ')
            if len(iban_and_bic) == 2:
                header['iban'], header['bic'] = iban_and_bic
            line_number += 1

        elif record_type in TO_FILE_RECORD_TYPES:
            record = parse_records(line, TO_FILE_RECORD, line_number=line_number)
            convert_date_fields(record, TO_FILE_RECORD_DATES, tz)
            convert_decimal_fields(record, TO_FILE_RECORD_DECIMALS)

            line_number += 1
            # check for record extra info
            if line_number <= nlines:
                line = lines[line_number-1]
                while line[:3] in TO_FILE_RECORD_EXTRA_INFO_TYPES:
                    parse_record_extra_info(record, line, line_number=line_number)
                    line_number += 1
                    line = lines[line_number-1]

            records.append(record)
        elif record_type == 'T40':
            balance = parse_records(line, TO_BALANCE_RECORD, line_number=line_number)
            convert_date_fields(balance, TO_BALANCE_RECORD_DATES, tz)
            convert_decimal_fields(balance, TO_BALANCE_RECORD_DECIMALS)
            line_number += 1
        elif record_type == 'T50':
            cumulative = parse_records(line, TO_CUMULATIVE_RECORD, line_number=line_number)
            convert_date_fields(cumulative, TO_CUMULATIVE_RECORD_DATES, tz)
            convert_decimal_fields(cumulative, TO_CUMULATIVE_RECORD_DECIMALS)
            line_number += 1
        elif record_type == 'T51':
            cumulative_adjustment = parse_records(line, TO_CUMULATIVE_RECORD, line_number=line_number)
            convert_date_fields(cumulative_adjustment, TO_CUMULATIVE_RECORD_DATES, tz)
            convert_decimal_fields(cumulative_adjustment, TO_CUMULATIVE_RECORD_DECIMALS)
            line_number += 1
        elif record_type == 'T60' or record_type == 'T70':
            special_records.append(parse_records(line, TO_SPECIAL_RECORD, line_number=line_number, check_record_length=False))
            line_number += 1
        else:
            raise ValidationError(_('Unknown record type on {}({}): {}').format(filename, line_number, record_type))

    statements.append(combine_statement(header, records, balance, cumulative, cumulative_adjustment, special_records))
    return statements


def parse_filename_suffix(filename: str) -> str:
    a = filename.rsplit('.', 1)
    return a[len(a)-1]


def parse_tiliote_statements_from_file(filename: str) -> dict:
    if parse_filename_suffix(filename).upper() not in ('TO', 'TXT'):
        raise ValidationError(_('Not "tiliote" (.TO) file') + ': {}'.format(filename))
    with open(filename, 'rt', encoding='ISO-8859-1') as fp:
        return parse_tiliote_statements(fp.read(), filename=basename(filename))


def combine_svm_batch(header: dict, records: list, summary: dict) -> dict:
    data = {'header': header, 'records': records}
    if summary is not None:
        data['summary'] = summary
    return data


def parse_svm_batches(content: str, filename: str) -> list:
    lines = content.split('\n')
    nlines = len(lines)

    line_number = 1
    tz = timezone('Europe/Helsinki')

    batches = []
    header = None
    summary = None
    records = []

    while line_number <= nlines:
        line = lines[line_number-1]
        if line.strip() == '':
            line_number += 1
            continue
        record_type = line[:1]

        if record_type in SVM_FILE_HEADER_TYPES:
            if header:
                batches.append(combine_svm_batch(header, records, summary))
                header, records, summary = None, [], None
            header = parse_records(lines[line_number - 1], SVM_FILE_HEADER, line_number=line_number)
            convert_date_fields(header, SVM_FILE_HEADER_DATES, tz)
            line_number += 1
        elif record_type in SVM_FILE_RECORD_TYPES:
            record = parse_records(line, SVM_FILE_RECORD, line_number=line_number)
            convert_date_fields(record, SVM_FILE_RECORD_DATES, tz)
            convert_decimal_fields(record, SVM_FILE_RECORD_DECIMALS)
            line_number += 1
            records.append(record)
        elif record_type in SVM_FILE_SUMMARY_TYPES:
            summary = parse_records(line, SVM_FILE_SUMMARY, line_number=line_number)
            convert_decimal_fields(summary, SVM_FILE_SUMMARY_DECIMALS)
            line_number += 1
        else:
            raise ValidationError(_('Unknown record type on {}({}): {}').format(filename, line_number, record_type))

    batches.append(combine_svm_batch(header, records, summary))
    return batches


def parse_svm_batches_from_file(filename: str) -> dict:
    if parse_filename_suffix(filename).upper() not in ('SVM', 'TXT', 'KTL'):
        raise ValidationError(_('Not "saapuvat viitemaksut" (.SVM) file') + ': {}'.format(filename))
    with open(filename, 'rt', encoding='ISO-8859-1') as fp:
        return parse_svm_batches(fp.read(), filename=basename(filename))



import logging
import os
import traceback
import getpass
from os.path import basename
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files.move import file_move_safe
from django.core.management import CommandParser
from jbank.files import list_dir_files
from jbank.models import Payout, PayoutStatus, PAYOUT_ERROR, PAYOUT_CANCELED, PAYOUT_WAITING_UPLOAD, PAYOUT_UPLOADED
from jbank.wsedi import wsedi_get, wsedi_upload_file
from jutil.command import SafeCommand
from jutil.email import send_email


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
        Upload Finnish bank files. Assumes WS-EDI API parameter compatible HTTP REST API end-point.
        Uses project settings WSEDI_URL and WSEDI_TOKEN.
        By default uploads files of Payouts in WAITING_UPLOAD state.
        """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--payout', type=int)
        parser.add_argument('--file-type', type=str, help='E.g. XL, NDCORPAYS, pain.001.001.03')
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--force', action='store_true')

    def do(self, *args, **options):
        file_type = options['file_type']
        if not file_type:
            return print('--file-type required (e.g. XL, NDCORPAYS, pain.001.001.03)')

        payouts = Payout.objects.all()
        if options['payout']:
            payouts = Payout.objects.filter(id=options['payout'])
        else:
            payouts = payouts.filter(state=PAYOUT_WAITING_UPLOAD)

        for p in list(payouts):
            assert isinstance(p, Payout)
            try:
                # upload file
                logger.info('Uploading payment id={} {} file {}'.format(p.id, file_type, p.full_path))
                with open(p.full_path, 'rt') as fp:
                    file_content = fp.read()
                p.state = PAYOUT_UPLOADED
                p.save(update_fields=['state'])
                res = wsedi_upload_file(file_content, file_type, p.file_name)
                logger.info('HTTP response {}'.format(res.status_code))
                logger.info(res.text)

                # parse response
                data = res.json()
                response_code = data.get('ResponseCode', '')[:4]
                response_text = data.get('ResponseText', '')[:255]
                fds = data.get("FileDescriptors", {}).get("FileDescriptor", [])
                fd = {} if len(fds) == 0 else fds[0]
                file_reference = fd.get('FileReference', '')
                if not file_reference:
                    raise Exception("FileReference missing from response")
                p.file_reference = file_reference
                p.save(update_fields=['file_reference'])
                PayoutStatus.objects.create(payout=p, msg_id=p.msg_id, file_name=p.file_name, response_code=response_code, response_text=response_text, status_reason='File upload OK')

            except Exception as e:
                long_err = "File upload failed ({}): ".format(p.full_path) + traceback.format_exc()
                logger.error(long_err)
                short_err = 'File upload failed: ' + str(e)
                p.state = PAYOUT_ERROR
                p.save(update_fields=['state'])
                PayoutStatus.objects.create(payout=p, msg_id=p.msg_id, file_name=p.file_name, response_code=response_code, response_text=response_text, status_reason=short_err[:255])



import logging
import os
from copy import copy
from pathlib import Path
from pprint import pprint
from django.conf import settings
from django.core.files import File
from django.core.management import CommandParser
from django.db import transaction
from django.utils import translation
from jacc.models import Account, AccountType
from jbank.helpers import create_statement, create_reference_payment_batch, get_or_create_bank_account
from jbank.files import list_dir_files
from jbank.models import Statement, ReferencePaymentBatch, ReferencePaymentBatchFile
from jbank.parsers import parse_svm_batches_from_file
from jutil.command import SafeCommand
from django.utils.translation import ugettext as _


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = 'Parses bank account statement .SVM (saapuvat viitemaksut) files'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('path', type=str)
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--test', action='store_true')
        parser.add_argument('--delete-old', action='store_true')
        parser.add_argument('--auto-create-accounts', action='store_true')

    def do(self, *args, **options):
        files = list_dir_files(options['path'])
        # pprint(files)
        for filename in files:
            plain_filename = os.path.basename(filename)
            if options['delete_old']:
                ReferencePaymentBatch.objects.filter(name=plain_filename).delete()

            if options['test']:
                batches = parse_svm_batches_from_file(filename)
                pprint(batches)
                continue

            if not ReferencePaymentBatch.objects.filter(name=plain_filename).first():
                logger.info('Importing reference payment batch file {}'.format(plain_filename))

                batches = parse_svm_batches_from_file(filename)
                if options['verbose']:
                    pprint(batches)

                with transaction.atomic():
                    if not ReferencePaymentBatch.objects.filter(name=plain_filename).first():
                        file = ReferencePaymentBatchFile()
                        file.save()
                        with open(filename, 'rb') as fp:
                            file.file.save(plain_filename, File(fp))

                        for data in batches:
                            if options['auto_create_accounts']:
                                for rec_data in data['records']:
                                    account_number = rec_data.get('account_number')
                                    if account_number:
                                        get_or_create_bank_account(account_number)

                            create_reference_payment_batch(data, name=plain_filename, file=file)
            else:
                print('Skipping reference payment batch file {}'.format(filename))



# Generated by Django 2.0 on 2018-02-09 09:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0020_payout_due_date'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payoutstatus',
            name='payout',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='payoutstatus_set', to='jbank.Payout', verbose_name='payout'),
        ),
    ]



# Generated by Django 2.0.4 on 2018-04-25 17:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0023_auto_20180419_1316'),
    ]

    operations = [
        migrations.AddField(
            model_name='payout',
            name='full_path',
            field=models.TextField(blank=True, editable=False, verbose_name='full path'),
        ),
        migrations.AlterField(
            model_name='payout',
            name='state',
            field=models.CharField(blank=True, choices=[('W', 'waiting processing'), ('U', 'waiting upload'), ('D', 'uploaded'), ('P', 'paid'), ('C', 'canceled'), ('E', 'error')], db_index=True, default='W', max_length=1, verbose_name='state'),
        ),
    ]



import logging
import os
import subprocess
import traceback
from decimal import Decimal
from os.path import join
from django.conf import settings
from django.core.management import CommandParser
from django.db import transaction
from jbank.models import Payout, PAYOUT_ERROR, PAYOUT_WAITING_PROCESSING, PayoutStatus, PAYOUT_WAITING_UPLOAD
from jbank.sepa import Pain001, PAIN001_REMITTANCE_INFO_MSG, PAIN001_REMITTANCE_INFO_OCR_ISO, \
    PAIN001_REMITTANCE_INFO_OCR
from jutil.command import SafeCommand
from jutil.format import format_xml
from jutil.validators import iban_bic


logger = logging.getLogger(__name__)


class Command(SafeCommand):
    help = """
        Generates pain.001.001.03 compatible SEPA payment files from pending Payout objects.
        By default generates files of Payouts in WAITING_PROCESSING state.
        """

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('dir', type=str)
        parser.add_argument('--payout', type=int)
        parser.add_argument('--verbose', action='store_true')

    def do(self, *args, **options):
        if options['verbose']:
            logger.info('Writing pain.001 files to {}'.format(options['dir']))

        payouts = Payout.objects.all()
        if options['payout']:
            payouts = Payout.objects.filter(id=options['payout'])
        else:
            payouts = payouts.filter(state=PAYOUT_WAITING_PROCESSING)

        for p in list(payouts):
            assert isinstance(p, Payout)
            try:
                if options['verbose']:
                    print(p)

                if not p.msg_id or not p.file_name:
                    p.generate_msg_id()
                    p.file_name = p.msg_id + '.XL'
                    p.save(update_fields=['msg_id', 'file_name'])

                pain001 = Pain001(p.msg_id, p.payer.name, p.payer.account_number, p.payer.bic, p.payer.org_id, p.payer.address_lines, p.payer.country_code)
                if p.messages:
                    remittance_info = p.messages
                    remittance_info_type = PAIN001_REMITTANCE_INFO_MSG
                else:
                    remittance_info = p.reference
                    remittance_info_type = PAIN001_REMITTANCE_INFO_OCR_ISO if remittance_info[:2] == 'RF' else PAIN001_REMITTANCE_INFO_OCR
                pain001.add_payment(p.id, p.recipient.name, p.recipient.account_number, p.recipient.bic, p.amount, remittance_info, remittance_info_type, p.due_date)

                p.full_path = full_path = os.path.join(options['dir'], p.file_name)
                if options['verbose']:
                    print(p, 'written to', full_path)
                pain001.render_to_file(full_path)
                logger.info('{} generated'.format(full_path))
                p.state = PAYOUT_WAITING_UPLOAD
                p.save(update_fields=['full_path', 'state'])

                PayoutStatus.objects.create(payout=p, file_name=p.file_name, msg_id=p.msg_id, status_reason="File generation OK")
            except Exception as e:
                short_err = 'File generation failed: ' + str(e)
                long_err = "File generation failed ({}): ".format(p.file_name) + traceback.format_exc()
                logger.error(long_err)
                p.state = PAYOUT_ERROR
                p.save(update_fields=['state'])
                PayoutStatus.objects.create(payout=p, file_name=p.file_name, msg_id=p.msg_id, status_reason=short_err[:255])



import logging
import re
from datetime import datetime, time, date
from os.path import basename, join
import pytz
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.db import models
from django.utils.text import format_lazy
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from jacc.models import AccountEntry, AccountEntrySourceFile, Account
from jutil.dict import choices_label
from jutil.validators import fi_iban_validator, iban_validator, iban_bic, iso_payment_reference_validator, \
    fi_payment_reference_validator

logger = logging.getLogger(__name__)


RECORD_ENTRY_TYPE = (
    ('1', _('Deposit')),
    ('2', _('Withdrawal')),
    ('3', _('Deposit Correction')),
    ('4', _('Withdrawal Correction')),
)

RECORD_CODES = (
    ('700', _('Money Transfer (In/Out)')),
    ('701', _('Recurring Payment (In/Out)')),
    ('702', _('Bill Payment (Out)')),
    ('703', _('Payment Terminal Deposit (In)')),
    ('704', _('Bank Draft (In/Out)')),
    ('705', _('Reference Payments (In)')),
    ('706', _('Payment Service (Out)')),
    ('710', _('Deposit (In)')),
    ('720', _('Withdrawal (Out)')),
    ('721', _('Card Payment (Out)')),
    ('722', _('Check (Out)')),
    ('730', _('Bank Fees (Out)')),
    ('740', _('Interests Charged (Out)')),
    ('750', _('Interests Credited (In)')),
    ('760', _('Loan (Out)')),
    ('761', _('Loan Payment (Out)')),
    ('770', _('Foreign Transfer (In/Out)')),
    ('780', _('Zero Balancing (In/Out)')),
    ('781', _('Sweeping (In/Out)')),
    ('782', _('Topping (In/Out)')),
)

RECEIPT_CODE = (
    ('', ''),
    ('E', _('Separate')),
    ('P', _('Separate/Paper')),
)

CURRENCY_IDENTIFIERS = (
    ('1', 'EUR'),
)

NAME_SOURCES = (
    ('', _('Not Set')),
    ('A', _('From Customer')),
    ('K', _('From Bank Clerk')),
    ('J', _('From Bank System')),
)

CORRECTION_IDENTIFIER = (
    ('0', _('Regular Entry')),
    ('1', _('Correction Entry')),
)

DELIVERY_METHOD = (
    ('A', _('From Customer')),
    ('K', _('From Bank Clerk')),
    ('J', _('From Bank System')),
)

PAYOUT_WAITING_PROCESSING = 'W'
PAYOUT_WAITING_UPLOAD = 'U'
PAYOUT_UPLOADED = 'D'
PAYOUT_PAID = 'P'
PAYOUT_CANCELED = 'C'
PAYOUT_ERROR = 'E'

PAYOUT_STATE = (
    (PAYOUT_WAITING_PROCESSING, _('waiting processing')),
    (PAYOUT_WAITING_UPLOAD, _('waiting upload')),
    (PAYOUT_UPLOADED, _('uploaded')),
    (PAYOUT_PAID, _('paid')),
    (PAYOUT_CANCELED, _('canceled')),
    (PAYOUT_ERROR, _('error')),
)


class Statement(AccountEntrySourceFile):
    file = models.ForeignKey('StatementFile', blank=True, default=None, null=True, on_delete=models.CASCADE)
    account = models.ForeignKey(Account, related_name='+', on_delete=models.PROTECT)
    account_number = models.CharField(_('account number'), max_length=32, db_index=True)
    statement_number = models.SmallIntegerField(_('statement number'), )
    begin_date = models.DateField(_('begin date'), db_index=True)
    end_date = models.DateField(_('end date'), db_index=True)
    record_date = models.DateTimeField(_('record date'), db_index=True)
    customer_identifier = models.CharField(_('customer identifier'), max_length=32)
    begin_balance_date = models.DateField(_('begin balance date'), )
    begin_balance = models.DecimalField(_('begin balance'), max_digits=10, decimal_places=2)
    record_count = models.IntegerField(_('record count'), null=True, default=None)
    currency_code = models.CharField(_('currency code'), max_length=3)
    account_name = models.CharField(_('account name'), max_length=32, blank=True, default='')
    account_limit = models.DecimalField(_('account limit'), max_digits=10, decimal_places=2, blank=True, default=None, null=True)
    owner_name = models.CharField(_('owner name'), max_length=64)
    contact_info_1 = models.CharField(_('contact info (1)'), max_length=64)
    contact_info_2 = models.CharField(_('contact info (2)'), max_length=64, blank=True, default='')
    bank_specific_info_1 = models.CharField(_('bank specific info (1)'), max_length=32, blank=True, default='')
    iban = models.CharField(_('IBAN'), max_length=32, db_index=True)
    bic = models.CharField(_('BIC'), max_length=8, db_index=True)

    class Meta:
        verbose_name = _('statement')
        verbose_name_plural = _('statements')


class PaymentRecordManager(models.Manager):
    def filter_matched(self):
        return self.exclude(child_set=None)

    def filter_unmatched(self):
        return self.filter(child_set=None)


class StatementRecord(AccountEntry):
    objects = PaymentRecordManager()
    statement = models.ForeignKey(Statement, verbose_name=_('statement'), related_name='record_set', on_delete=models.CASCADE)
    record_number = models.IntegerField(_('record number'))
    archive_identifier = models.CharField(_('archive identifier'), max_length=32, blank=True, default='', db_index=True)
    record_date = models.DateField(_('record date'), db_index=True)
    value_date = models.DateField(_('value date'), db_index=True, blank=True, null=True, default=None)
    paid_date = models.DateField(_('paid date'), db_index=True, blank=True, null=True, default=None)
    entry_type = models.CharField(_('entry type'), max_length=1, choices=RECORD_ENTRY_TYPE, db_index=True)
    record_code = models.CharField(_('record type'), max_length=3, choices=RECORD_CODES, db_index=True)
    record_description = models.CharField(_('record description'), max_length=64)
    receipt_code = models.CharField(_('receipt code'), max_length=1, choices=RECEIPT_CODE, db_index=True, blank=True)
    delivery_method = models.CharField(_('delivery method'), max_length=1, db_index=True, choices=DELIVERY_METHOD)
    name = models.CharField(_('name'), max_length=64, blank=True, db_index=True)
    name_source = models.CharField(_('name source'), max_length=1, blank=True, choices=NAME_SOURCES)
    recipient_account_number = models.CharField(_('recipient account number'), max_length=32, blank=True, db_index=True)
    recipient_account_number_changed = models.CharField(_('recipient account number changed'), max_length=1, blank=True)
    remittance_info = models.CharField(_('remittance info'), max_length=32, db_index=True, blank=True)
    messages = models.TextField(_('messages'), blank=True, default='')
    client_messages = models.TextField(_('client messages'), blank=True, default='')
    bank_messages = models.TextField(_('bank messages'), blank=True, default='')

    class Meta:
        verbose_name = _('statement record')
        verbose_name_plural = _('statement records')

    def clean(self):
        self.source_file = self.statement
        paid_date = self.paid_date
        assert isinstance(paid_date, date)
        self.timestamp = pytz.utc.localize(datetime.combine(paid_date, time(0, 0)))
        self.description = '{name}: {record_description}'.format(record_description=self.record_description, name=self.name)


class StatementRecordSepaInfo(models.Model):
    record = models.OneToOneField(StatementRecord, verbose_name=_('record'), related_name='sepa_info', on_delete=models.CASCADE)
    reference = models.CharField(_('reference'), max_length=35, blank=True)
    iban_account_number = models.CharField(_('IBAN'), max_length=35, blank=True)
    bic_code = models.CharField(_('BIC'), max_length=35, blank=True)
    recipient_name_detail = models.CharField(_('recipient name detail'), max_length=70, blank=True)
    payer_name_detail = models.CharField(_('payer name detail'), max_length=70, blank=True)
    identifier = models.CharField(_('identifier'), max_length=35, blank=True)
    archive_identifier = models.CharField(_('archive identifier'), max_length=35, blank=True)

    class Meta:
        verbose_name = _('SEPA')
        verbose_name_plural = _('SEPA')

    def __str__(self):
        return '[{}]'.format(self.id)


class ReferencePaymentBatchManager(models.Manager):
    def latest_record_date(self) -> datetime:
        """
        :return: datetime of latest record available or None
        """
        obj = self.order_by('-record_date').first()
        if not obj:
            return None
        return obj.record_date


class ReferencePaymentBatch(AccountEntrySourceFile):
    objects = ReferencePaymentBatchManager()
    file = models.ForeignKey('ReferencePaymentBatchFile', blank=True, default=None, null=True, on_delete=models.CASCADE)
    record_date = models.DateTimeField(_('record date'), db_index=True)
    institution_identifier = models.CharField(_('institution identifier'), max_length=2)
    service_identifier = models.CharField(_('service identifier'), max_length=9, blank=True)
    currency_identifier = models.CharField(_('currency identifier'), max_length=3, choices=CURRENCY_IDENTIFIERS)

    class Meta:
        verbose_name = _('reference payment batch')
        verbose_name_plural = _('reference payment batches')


class ReferencePaymentRecord(AccountEntry):
    """
    Reference payment record. See jacc.Invoice for date/time variable naming conventions.
    """
    objects = PaymentRecordManager()
    batch = models.ForeignKey(ReferencePaymentBatch, verbose_name=_('batch'), related_name='record_set', on_delete=models.CASCADE)
    record_type = models.CharField(_('record type'), max_length=1)
    account_number = models.CharField(_('account number'), max_length=32, db_index=True)
    record_date = models.DateField(_('record date'), db_index=True)
    paid_date = models.DateField(_('paid date'), db_index=True, blank=True, null=True, default=None)
    archive_identifier = models.CharField(_('archive identifier'), max_length=32, blank=True, default='', db_index=True)
    remittance_info = models.CharField(_('remittance info'), max_length=32, db_index=True)
    payer_name = models.CharField(_('payer name'), max_length=12, db_index=True)
    currency_identifier = models.CharField(_('currency identifier'), max_length=1, choices=CURRENCY_IDENTIFIERS)
    name_source = models.CharField(_('name source'), max_length=1, choices=NAME_SOURCES, blank=True)
    correction_identifier = models.CharField(_('correction identifier'), max_length=1, choices=CORRECTION_IDENTIFIER)
    delivery_method = models.CharField(_('delivery method'), max_length=1, db_index=True, choices=DELIVERY_METHOD)
    receipt_code = models.CharField(_('receipt code'), max_length=1, choices=RECEIPT_CODE, db_index=True, blank=True)

    class Meta:
        verbose_name = _('reference payment records')
        verbose_name_plural = _('reference payment records')

    @property
    def remittance_info_short(self) -> str:
        """
        Remittance info without preceding zeroes.
        :return: str
        """
        return re.sub(r'^0+', '', self.remittance_info)

    def clean(self):
        self.source_file = self.batch
        paid_date = self.paid_date
        assert isinstance(paid_date, date)
        self.timestamp = pytz.utc.localize(datetime.combine(paid_date, time(0, 0)))
        self.description = '{amount} {remittance_info} {payer_name}'.format(amount=self.amount, remittance_info=self.remittance_info, payer_name=self.payer_name)


class StatementFile(models.Model):
    created = models.DateTimeField(_('created'), default=now, db_index=True, blank=True, editable=False)
    file = models.FileField(_('file'), upload_to='uploads')
    errors = models.TextField(_('errors'), max_length=4086, default='', blank=True)

    class Meta:
        verbose_name = _('account statement file')
        verbose_name_plural = _('account statement files')

    @property
    def full_path(self):
        return join(settings.BASE_DIR, self.file.name) if self.file else ''

    def __str__(self):
        return basename(str(self.file.name)) if self.file else ''


class ReferencePaymentBatchFile(models.Model):
    created = models.DateTimeField(_('created'), default=now, db_index=True, blank=True, editable=False)
    file = models.FileField(_('file'), upload_to='uploads')
    errors = models.TextField(_('errors'), max_length=4086, default='', blank=True)

    class Meta:
        verbose_name = _("reference payment batch file")
        verbose_name_plural = _("reference payment batch files")

    @property
    def full_path(self):
        return join(settings.BASE_DIR, self.file.name) if self.file else ''

    def __str__(self):
        return basename(str(self.file.name)) if self.file else ''


class PayoutParty(models.Model):
    name = models.CharField(_('name'), max_length=128, db_index=True)
    account_number = models.CharField(_('account number'), max_length=35, db_index=True, validators=[iban_validator])
    bic = models.CharField(_('BIC'), max_length=16, db_index=True, blank=True)
    org_id = models.CharField(_('organization id'), max_length=32, db_index=True, blank=True, default='')
    address = models.TextField(_('address'), blank=True, default='')
    country_code = models.CharField(_('country code'), max_length=2, default='FI', blank=True, db_index=True)
    payouts_account = models.ForeignKey(Account, verbose_name=_('payouts account'), null=True, default=None, blank=True, on_delete=models.PROTECT)

    class Meta:
        verbose_name = _("payout party")
        verbose_name_plural = _("payout parties")

    def __str__(self):
        return self.name

    def clean(self):
        if not self.bic:
            self.bic = iban_bic(self.account_number)

    @property
    def address_lines(self):
        out = []
        for line in self.address.split('\n'):
            line = line.strip()
            if line:
                out.append(line)
        return out


class Payout(AccountEntry):
    payer = models.ForeignKey(PayoutParty, verbose_name=_('payer'), related_name='+', on_delete=models.PROTECT)
    recipient = models.ForeignKey(PayoutParty, verbose_name=_('recipient'), related_name='+', on_delete=models.PROTECT)
    messages = models.TextField(_('recipient messages'), blank=True, default='')
    reference = models.CharField(_('recipient reference'), blank=True, default='', max_length=32)
    msg_id = models.CharField(_('message id'), max_length=64, blank=True, db_index=True, editable=False)
    file_name = models.CharField(_('file name'), max_length=255, blank=True, db_index=True, editable=False)
    full_path = models.TextField(_('full path'), blank=True, editable=False)
    file_reference = models.CharField(_('file reference'), max_length=255, blank=True, db_index=True, editable=False)
    due_date = models.DateField(_('due date'), db_index=True, blank=True, null=True, default=None)
    paid_date = models.DateTimeField(_('paid date'), db_index=True, blank=True, null=True, default=None)
    state = models.CharField(_('state'), max_length=1, blank=True, default=PAYOUT_WAITING_PROCESSING, choices=PAYOUT_STATE, db_index=True)

    class Meta:
        verbose_name = _("payout")
        verbose_name_plural = _("payouts")

    def clean(self):
        if self.parent and not self.amount:
            self.amount = self.parent.amount

        # prevent defining both reference and messages
        if self.messages and self.reference:
            raise ValidationError(_('Both recipient messages and recipient reference cannot be defined simultenously'))

        # validate reference if any
        if self.reference:
            if self.reference[:2] == 'RF':
                iso_payment_reference_validator(self.reference)
            else:
                fi_payment_reference_validator(self.reference)

        # prevent canceling payouts which have been uploaded successfully
        if self.state == PAYOUT_CANCELED:
            if self.is_upload_done:
                group_status = self.group_status
                if group_status != 'RJCT':
                    raise ValidationError(_('File already uploaded') + ' ({})'.format(group_status))

    def generate_msg_id(self):
        self.msg_id = re.sub(r'[^\d]', '', now().date().isoformat()) + 'P' + str(self.id)

    @property
    def state_name(self):
        return choices_label(PAYOUT_STATE, self.state)

    @property
    def is_upload_done(self):
        return PayoutStatus.objects.filter(payout=self, response_code='00').first() is not None

    @property
    def is_accepted(self):
        return self.group_status == 'ACCP'

    @property
    def is_rejected(self):
        return self.group_status == 'RJCT'

    @property
    def group_status(self):
        status = PayoutStatus.objects.filter(payout=self).order_by('-id').first()
        return status.group_status if status else ''


class PayoutStatusManager(models.Manager):
    def is_file_processed(self, filename: str) -> bool:
        return self.filter(file_name=basename(filename)).first() is not None


class PayoutStatus(models.Model):
    objects = PayoutStatusManager()
    payout = models.ForeignKey(Payout, verbose_name=_('payout'), related_name='payoutstatus_set', on_delete=models.PROTECT, null=True, default=None, blank=True)
    created = models.DateTimeField(_('created'), default=now, db_index=True, editable=False, blank=True)
    file_name = models.CharField(_('file name'), max_length=255, blank=True, db_index=True, editable=False)
    response_code = models.CharField(_('response code'), max_length=4, blank=True, db_index=True)
    response_text = models.CharField(_('response text'), max_length=128, blank=True)
    msg_id = models.CharField(_('message id'), max_length=64, blank=True, db_index=True)
    original_msg_id = models.CharField(_('original message id'), blank=True, max_length=64, db_index=True)
    group_status = models.CharField(_('group status'), max_length=8, blank=True, db_index=True)
    status_reason = models.CharField(_('status reason'), max_length=255, blank=True)

    class Meta:
        verbose_name = _("payout status")
        verbose_name_plural = _("payout statuses")

    def __str__(self):
        return self.group_status

    @property
    def is_accepted(self):
        return self.group_status == 'ACCP'

    @property
    def is_rejected(self):
        return self.group_status == 'RJCT'


class Refund(Payout):
    class Meta:
        verbose_name = _("refund")
        verbose_name_plural = _("refunds")



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-11-02 23:41
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0005_statementfile_created'),
    ]

    operations = [
        migrations.AddField(
            model_name='statementfile',
            name='errors',
            field=models.TextField(blank=True, default='', max_length=4086),
        ),
        migrations.AlterField(
            model_name='statementfile',
            name='file',
            field=models.FileField(upload_to='uploads', verbose_name='file'),
        ),
    ]



# Generated by Django 2.0 on 2018-01-26 08:58

from django.db import migrations, models
import django.db.models.deletion
import jutil.validators


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0008_auto_20180121_1705'),
        ('jbank', '0011_auto_20180126_0851'),
    ]

    operations = [
        migrations.CreateModel(
            name='RefundRecord',
            fields=[
                ('accountentry_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='jacc.AccountEntry')),
                ('iban_account_number', models.CharField(max_length=35, validators=[jutil.validators.iban_validator], verbose_name='IBAN')),
                ('refund_date', models.DateField(blank=True, db_index=True, default=None, null=True, verbose_name='refund date')),
                ('archive_identifier', models.CharField(blank=True, db_index=True, default='', max_length=32, verbose_name='archive identifier')),
                ('state', models.CharField(blank=True, db_index=True, default='W', max_length=1, verbose_name='state')),
            ],
            options={
                'verbose_name_plural': 'refunds',
                'verbose_name': 'refund',
            },
            bases=('jacc.accountentry',),
        ),
        migrations.RemoveField(
            model_name='refund',
            name='source',
        ),
        migrations.DeleteModel(
            name='Refund',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-07 18:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0008_auto_20171103_0007'),
    ]

    operations = [
        migrations.AlterField(
            model_name='statementrecord',
            name='remittance_info',
            field=models.CharField(blank=True, db_index=True, max_length=32, verbose_name='remittance info'),
        ),
    ]



# Generated by Django 2.0 on 2018-02-08 07:24

from django.db import migrations, models
import django.db.models.deletion
import jutil.validators


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0015_auto_20180208_0643'),
    ]

    operations = [
        migrations.CreateModel(
            name='PayoutParty',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=128, verbose_name='name')),
                ('account_number', models.CharField(db_index=True, max_length=35, validators=[jutil.validators.iban_validator], verbose_name='account number')),
                ('bic', models.CharField(blank=True, db_index=True, max_length=16, verbose_name='BIC')),
                ('org_id', models.CharField(blank=True, db_index=True, default='', max_length=32, verbose_name='organization id')),
                ('address', models.TextField(blank=True, default='', verbose_name='address')),
                ('country_code', models.CharField(blank=True, db_index=True, default='FI', max_length=2, verbose_name='country code')),
            ],
            options={
                'verbose_name_plural': 'payout parties',
                'verbose_name': 'payout party',
            },
        ),
        migrations.AlterModelOptions(
            name='payoutstatus',
            options={'verbose_name': 'payout status', 'verbose_name_plural': 'payout statuses'},
        ),
        migrations.RemoveField(
            model_name='payout',
            name='recipient_account_number',
        ),
        migrations.AlterField(
            model_name='payout',
            name='state',
            field=models.CharField(blank=True, choices=[('W', 'waiting'), ('P', 'paid'), ('C', 'canceled'), ('E', 'error')], db_index=True, default='W', max_length=1, verbose_name='state'),
        ),
        migrations.AddField(
            model_name='payout',
            name='payer',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='jbank.PayoutParty', verbose_name='payer'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='payout',
            name='recipient',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='jbank.PayoutParty', verbose_name='recipient'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.0.2 on 2018-04-11 18:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0021_auto_20180209_0935'),
    ]

    operations = [
        migrations.AlterField(
            model_name='referencepaymentbatch',
            name='service_identifier',
            field=models.CharField(blank=True, max_length=9, verbose_name='service identifier'),
        ),
    ]
